"""Tests around instance-linking functionality"""

import itertools
import logging
import os
import re
from collections import defaultdict
from pathlib import Path

import pytest

from enderchest import ShulkerBox
from enderchest import filesystem as fs
from enderchest import inventory, place
from enderchest import shulker_box as sb
from enderchest.loggers import IMPORTANT

from . import utils


class TestRglob:
    def test_max_depth_of_one_is_equivalent_to_a_plain_glob(self, minecraft_root):
        glob = sorted(minecraft_root.iterdir())
        assert len(glob) > 0  # meta-test

        assert glob == sorted(place._rglob(minecraft_root, 1))

    def test_max_depth_of_zero_finds_all_files_in_path(self, minecraft_root):
        search_dir = minecraft_root / "workspace"  # choose someplace with no links
        rglob = sorted([path for path in search_dir.rglob("*") if not path.is_dir()])
        assert len(rglob) > 0  # meta-test

        assert rglob == sorted(place._rglob(minecraft_root / "workspace", 0))

    def test_max_depth_of_two_returns_subdirectories(self, minecraft_root) -> None:
        instances_folder = minecraft_root / "instances"
        expected: list[Path] = [instances_folder / "instgroups.json"]
        for instance in utils.TESTING_INSTANCES[1:]:
            expected.extend(
                (
                    minecraft_root / instance.root,
                    minecraft_root / instance.root.parent / "instance.cfg",
                    minecraft_root / instance.root.parent / "mmc-pack.json",
                )
            )
        expected.sort()

        assert expected == sorted(place._rglob(instances_folder, 2))


class TestSingleShulkerPlace:
    """Test the simplest case of linking--where the files in the shulker should
    go into every instance"""

    @pytest.fixture(autouse=True)
    def setup_teardown(self, minecraft_root, home):
        """Setup / teardown for this test class"""
        chest_folder = minecraft_root / "EnderChest"
        utils.pre_populate_enderchest(chest_folder, utils.GLOBAL_SHULKER)

        do_not_touch = {
            (chest_folder / "global" / "logs" / "bumpona.log"): (
                "Like a bump on a bump on a log, baby.\n"
                "Like I'm in a fist fight with a fog, baby.\n"
                "Step-ball-change and a pirouette.\n"
                "And I regret, I regret.\n"
            ),
        }

        for path, contents in do_not_touch.items():
            path.parent.mkdir(parents=True, exist_ok=True)
            path.write_text(contents)

        yield

        # check on teardown that all those "do_not_touch" files are untouched
        for path, contents in do_not_touch.items():
            assert path.read_text() == contents

    def test_place_exits_when_there_is_no_enderchest(self, minecraft_root, caplog):
        placements = place.place_ender_chest(minecraft_root.parent)
        assert len(caplog.records) == 1
        assert (
            caplog.records[0].levelno,
            caplog.records[0].message.split(" from")[0],  # hacky "startswith"
        ) == (logging.ERROR, "Could not load EnderChest")

        assert placements == {}

    @pytest.mark.parametrize(
        "error_handling",
        ("abort", "ignore", "skip", "skip-instance", "skip-shulker-box"),
    )
    def test_place_handles_error_when_an_instance_is_missing(
        self, minecraft_root, error_handling, caplog
    ):
        instance_folder = minecraft_root / "instances" / "axolotl"
        safe_keeping = minecraft_root / "axolotl.bkp"

        instance_folder.rename(safe_keeping)

        try:
            placements = place.place_ender_chest(
                minecraft_root, error_handling=error_handling
            )
        finally:
            safe_keeping.rename(instance_folder)

        errors = [record for record in caplog.records if record.levelname == "ERROR"]
        assert errors[0].msg.startswith("No minecraft instance exists at")
        assert placements.get("axolotl", {}) == {}

    @pytest.mark.parametrize("relative", (True, False), ids=("relative", "absolute"))
    @utils.parametrize_over_instances("official", "axolotl")
    def test_respects_the_relative_parameter(self, minecraft_root, instance, relative):
        placements = place.place_ender_chest(minecraft_root, relative=relative)
        instance_folder = utils.resolve(instance.root, minecraft_root)

        assert placements[instance.name][Path("logs")] == ["global"]
        assert (instance_folder / "logs").readlink().is_absolute() is (not relative)

    @utils.parametrize_over_instances("official", "axolotl")
    def test_place_replaces_empty_folders(self, minecraft_root, instance):
        placements = place.place_ender_chest(minecraft_root)
        instance_folder = utils.resolve(instance.root, minecraft_root)

        assert placements[instance.name][Path("logs")] == ["global"]

        # and to be sure
        assert (instance_folder / "logs").resolve() == (
            minecraft_root / "EnderChest" / "global" / "logs"
        ).resolve()

        # finally, just to be explicit
        assert (instance_folder / "logs" / "bumpona.log").read_text(
            "utf-8"
        ).splitlines()[0] == "Like a bump on a bump on a log, baby."

    @utils.parametrize_over_instances("official", "axolotl")
    def test_place_is_able_to_place_files(self, minecraft_root, instance):
        # including in directories that didn't previously exist!
        placements = place.place_ender_chest(minecraft_root)

        instance_folder = utils.resolve(instance.root, minecraft_root)

        assert not (instance_folder / "config").is_symlink()

        resource_path = Path("config") / "iris.properties"
        assert placements[instance.name][resource_path] == ["global"]

        # and to be sure
        assert (instance_folder / resource_path).resolve() == (
            minecraft_root / "EnderChest" / "global" / resource_path
        )

    @utils.parametrize_over_instances("official", "axolotl")
    def test_place_doesnt_choke_on_relative_root(
        self, minecraft_root, instance, monkeypatch
    ):
        monkeypatch.chdir(minecraft_root.parent)
        placements = place.place_ender_chest(Path(minecraft_root.name))

        instance_folder = utils.resolve(instance.root, minecraft_root)

        assert not (instance_folder / "config").is_symlink()

        resource_path = Path("config") / "iris.properties"
        assert placements[instance.name][resource_path] == ["global"]

        # and to be sure
        assert (instance_folder / resource_path).resolve() == (
            minecraft_root / "EnderChest" / "global" / resource_path
        )

    @utils.parametrize_over_instances("official", "axolotl")
    def test_place_is_able_to_place_root_level_files(self, minecraft_root, instance):
        placements = place.place_ender_chest(minecraft_root)

        instance_folder = utils.resolve(instance.root, minecraft_root)

        assert placements[instance.name][Path("usercache.json")] == ["global"]

        # and to be sure
        assert (instance_folder / "usercache.json").resolve() == (
            minecraft_root / "EnderChest" / "global" / "usercache.json"
        )

        # finally, just to be explicit
        assert (
            instance_folder / "usercache.json"
        ).read_text() == "alexander\nmomoa\nbateman\n"

    @utils.parametrize_over_instances("official", "axolotl")
    def test_does_not_place_shulker_config(self, minecraft_root, instance):
        placements = place.place_ender_chest(minecraft_root)

        instance_folder = utils.resolve(instance.root, minecraft_root)

        assert placements[instance.name].get(Path(fs.SHULKER_BOX_CONFIG_NAME), []) == []

        # and to be sure
        assert not (instance_folder / fs.SHULKER_BOX_CONFIG_NAME).exists()

    @utils.parametrize_over_instances("bee")
    def test_shulker_configs_can_be_placed_if_you_really_want(
        self, minecraft_root, instance
    ):
        # it's explicitly overridden in the forge config
        utils.pre_populate_enderchest(
            minecraft_root / "EnderChest", utils.OPTIFINE_SHULKER
        )
        placements = place.place_ender_chest(minecraft_root)

        instance_folder = utils.resolve(instance.root, minecraft_root)

        assert placements[instance.name][Path(fs.SHULKER_BOX_CONFIG_NAME)] == [
            "optifine"
        ]

        # and to be sure
        assert (
            instance_folder / fs.SHULKER_BOX_CONFIG_NAME
        ).resolve() == fs.shulker_box_config(minecraft_root, "optifine")

    @utils.parametrize_over_instances("official", "axolotl")
    def test_link_folder_can_be_a_symlink(self, minecraft_root, instance):
        placements = place.place_ender_chest(minecraft_root)

        instance_folder = utils.resolve(instance.root, minecraft_root)

        assert placements[instance.name][Path("crash-reports")] == ["global"]
        assert Path("crash-reports") / "20230524.log" not in placements[instance.name]

        # the counterpoint to the whole "one assertion per test" rule--this
        # is a cascading case of figuring way of figuring out just how badly
        # the impl is borked
        assert (instance_folder / "crash-reports").exists()
        assert (instance_folder / "crash-reports" / "20230524.log").read_text() == (
            "ERROR: Everything is broken\n" "WARNING: And somehow also on fire\n"
        )

        assert (instance_folder / "crash-reports").is_symlink()

        assert (instance_folder / "crash-reports" / "20230524.log").resolve() == (
            minecraft_root / "crash-reports" / "20230524.log"
        )

    @pytest.mark.parametrize("link_type", ("absolute", "relative"))
    @utils.parametrize_over_instances("official", "axolotl")
    def test_place_places_symlinks(self, minecraft_root, instance, link_type):
        placements = place.place_ender_chest(
            minecraft_root, relative=link_type == "relative"
        )

        instance_folder = utils.resolve(instance.root, minecraft_root)
        save_path = Path("saves") / "test"

        assert placements[instance.name][save_path] == ["global"]
        assert save_path / "level.dat" not in placements[instance.name]

        # the counterpoint to the whole "one assertion per test" rule--this
        # is a cascading case of figuring way of figuring out just how badly
        # the impl is borked

        assert (instance_folder / save_path).exists()
        assert (instance_folder / save_path / "level.dat").exists()
        assert (
            instance_folder / save_path / "level.dat"
        ).read_text() == "hello world\n"

        assert (instance_folder / save_path).is_symlink()
        assert not (instance_folder / save_path / "level.dat").is_symlink()

        assert (instance_folder / save_path / "level.dat").resolve() == (
            minecraft_root / "worlds" / "testbench" / "level.dat"
        )

    @utils.parametrize_over_instances("official", "axolotl")
    def test_absolute_symlinks_fully_resolve_target(self, minecraft_root, instance):
        placements = place.place_ender_chest(minecraft_root, relative=False)

        instance_folder = utils.resolve(instance.root, minecraft_root)
        save_path = Path("saves") / "test"

        link_target = os.path.abspath(os.readlink(instance_folder / save_path))

        assert placements[instance.name][save_path] == ["global"]
        assert save_path / "level.dat" not in placements[instance.name]

        # Windows shenanigans: https://bugs.python.org/issue42957
        if link_target.startswith(("\\\\?\\", "\\??\\")):  # pragma: no cover
            link_target = link_target[4:]

        assert link_target == str(minecraft_root / "worlds" / "testbench")

    @utils.parametrize_over_instances("axolotl", "bee")
    def test_place_cleans_up_broken_symlinks_by_default(self, minecraft_root, instance):
        instance_folder = utils.resolve(instance.root, minecraft_root)
        broken_link = instance_folder / "shaderpacks" / "Seuss CitH.zip.txt"
        broken_link.symlink_to(minecraft_root / "i-do-not-exist.txt")
        assert broken_link in broken_link.parent.iterdir()

        _ = place.place_ender_chest(minecraft_root)

        assert broken_link not in broken_link.parent.iterdir()

    @utils.parametrize_over_instances("axolotl")
    def test_place_can_be_told_to_leave_broken_links_alone(
        self, minecraft_root, instance
    ):
        instance_folder = utils.resolve(instance.root, minecraft_root)
        broken_link = instance_folder / "shaderpacks" / "Seuss CitH.zip.txt"
        broken_link.symlink_to(minecraft_root / "i-do-not-exist.txt")
        assert broken_link in broken_link.parent.iterdir()

        _ = place.place_ender_chest(minecraft_root, keep_broken_links=True)

        assert broken_link in broken_link.parent.iterdir()

    @pytest.mark.parametrize("link_type", ("absolute", "relative"))
    @utils.parametrize_over_instances("official", "axolotl")
    def test_place_cleans_up_stale_symlinks_by_default(
        self, minecraft_root, instance, link_type
    ):
        instance_folder = utils.resolve(instance.root, minecraft_root)
        old_box = fs.shulker_box_root(minecraft_root, "Old Man Shulker")
        old_box.mkdir()
        old_file = old_box / "i-do-exist.txt"
        old_file.write_text("Hello there\n")
        # no config file means it's not actually / no longer a shulker box

        stale_link = instance_folder / "old-file.txt"
        if link_type == "absolute":
            stale_link.symlink_to(old_file)
        else:
            stale_link.symlink_to(os.path.relpath(old_file, stale_link.parent))

        _ = place.place_ender_chest(minecraft_root)

        assert stale_link not in stale_link.parent.iterdir()

        # but make sure the original file is okay
        assert old_file.read_text() == "Hello there\n"

    @pytest.mark.parametrize("link_type", ("absolute", "relative"))
    @utils.parametrize_over_instances("official", "axolotl")
    def test_place_will_not_remove_links_pointing_outside_of_enderchest(
        self,
        minecraft_root,
        instance,
        link_type,
    ):
        instance_folder = utils.resolve(instance.root, minecraft_root)
        old_file = minecraft_root / "workspace" / "i-do-exist.txt"
        old_file.write_text("Hello there\n")

        stale_link = instance_folder / "old_file.txt"
        if link_type == "absolute":
            stale_link.symlink_to(old_file)
        else:
            stale_link.symlink_to(os.path.relpath(old_file, stale_link.parent))

        _ = place.place_ender_chest(minecraft_root)

        assert stale_link in stale_link.parent.iterdir()

    @pytest.mark.parametrize("link_type", ("absolute", "relative"))
    @utils.parametrize_over_instances("official", "axolotl")
    def test_stale_link_cleaning_is_based_on_direct_target(
        self,
        minecraft_root,
        instance,
        link_type,
    ):
        instance_folder = utils.resolve(instance.root, minecraft_root)
        working_file = minecraft_root / "workspace" / "i-do-exist.txt"
        working_file.write_text("Hello there\n")

        box_link = fs.shulker_box_root(minecraft_root, "some_box") / "valid.txt"
        box_link.parent.mkdir()
        box_link.symlink_to(working_file)

        link_link = instance_folder / "valid.txt"
        if link_type == "absolute":
            link_link.symlink_to(box_link)
        else:
            link_link.symlink_to(os.path.relpath(box_link, link_link.parent))

        _ = place.place_ender_chest(minecraft_root)

        assert link_link not in link_link.parent.iterdir()

        # and then just make sure that the originals are okay
        assert working_file.read_text() == "Hello there\n"
        assert box_link.resolve() == working_file

    @pytest.mark.parametrize("link_type", ("absolute", "relative"))
    @utils.parametrize_over_instances("official", "axolotl")
    def test_place_can_be_told_to_leave_stale_links_alone(
        self,
        minecraft_root,
        instance,
        link_type,
    ):
        instance_folder = utils.resolve(instance.root, minecraft_root)
        old_box = fs.shulker_box_root(minecraft_root, "Old Man Shulker")
        old_box.mkdir()
        old_file = old_box / "i-do-exist.txt"
        old_file.write_text("Hello there")

        stale_link = instance_folder / "old-file.txt"
        if link_type == "absolute":
            stale_link.symlink_to(old_file)
        else:
            stale_link.symlink_to(os.path.relpath(old_file, stale_link.parent))

        _ = place.place_ender_chest(minecraft_root, keep_stale_links=True)

        assert stale_link in stale_link.parent.iterdir()

    @utils.parametrize_over_instances("official", "axolotl")
    @pytest.mark.parametrize(
        "error_handling",
        ("abort", "ignore", "skip", "skip-instance", "skip-shulker-box"),
    )
    def test_place_will_not_overwrite_a_non_empty_folder(
        self,
        minecraft_root,
        instance,
        caplog,
        error_handling,
    ):
        instance_folder = utils.resolve(instance.root, minecraft_root)
        resource_path = Path("screenshots") / "thumbs.db"
        existing_file = instance_folder / resource_path
        existing_file.write_text("opposable")

        placements = place.place_ender_chest(
            minecraft_root, error_handling=error_handling
        )
        assert placements[instance.name].get(resource_path, []) == []

        error_log = "\n".join(
            record.getMessage()
            for record in caplog.records
            if record.levelno == logging.ERROR
        )
        assert re.search(
            rf"{instance.name}((.|\n)*)screenshots((.|\n)*)empty", error_log
        )

        # make sure the file is still there afterward
        assert existing_file.exists()
        assert existing_file.resolve() == existing_file
        assert existing_file.read_text() == "opposable"

    @utils.parametrize_over_instances("official", "axolotl")
    def test_place_will_not_overwrite_a_file(self, minecraft_root, instance, caplog):
        instance_folder = utils.resolve(instance.root, minecraft_root)
        resource_path = Path("usercache.json")
        existing_file = instance_folder / resource_path
        existing_file.write_text("isaacs\n")

        placements = place.place_ender_chest(minecraft_root)
        assert placements[instance.name].get(resource_path, []) == []

        error_log = "\n".join(
            record.getMessage()
            for record in caplog.records
            if record.levelname == "ERROR"
        )
        assert re.search(
            rf"{instance.name}((.|\n)*)usercache.json((.|\n)*)exists", error_log
        )

        # make sure the file is still there afterward
        assert existing_file.exists()
        assert existing_file.resolve() == existing_file
        assert existing_file.read_text() == "isaacs\n"

    @utils.parametrize_over_instances("official", "axolotl")
    def test_place_will_overwrite_an_existing_symlink(self, minecraft_root, instance):
        # this also tests that place will place files inside of a symlinked folder
        instance_folder = utils.resolve(instance.root, minecraft_root)
        original_target = minecraft_root / "workspace" / "teavsrp_lite.zip"
        original_target.write_text("I will trade you these\nfor less of these\n")
        resource_path = Path("resourcepacks") / "TEAVSRP.zip"
        existing_symlink = instance_folder / resource_path
        existing_symlink.symlink_to(original_target)

        placements = place.place_ender_chest(minecraft_root)
        assert placements[instance.name][resource_path] == ["global"]

        assert existing_symlink.read_text() == "Breaking News!\n"
        assert (
            existing_symlink.resolve()
            == (minecraft_root / "EnderChest" / "global" / resource_path).resolve()
        )

        # also make sure the original file is okay
        assert (
            original_target.read_text() == "I will trade you these\nfor less of these\n"
        )


class TestMatchesVersion:
    @pytest.mark.parametrize(
        "version", ("1.19.4", "1.20-pre6", "23w13a_or_b", "not even trying")
    )
    def test_simple_equality_checks(self, version):
        assert sb._matches_version(version, version)

    @pytest.mark.parametrize(
        "version_spec",
        ("1.19", "1.19.*", ">=1.19.0,<1.20"),
        ids=("minor-version", "wildcard", "bounding"),
    )
    def test_version_bounding(self, version_spec):
        assert sb._matches_version(version_spec, "1.19.4")

    @pytest.mark.parametrize(
        "version", ("1.19.4", "1.20-pre6", "23w13a_or_b", "not even trying")
    )
    def test_star_matches_anything(self, version):
        assert sb._matches_version("*", version)

    @pytest.mark.parametrize(
        "version", ("1.20-pre1", "1.20-rc1", "1.20.3-pre16", "1.20-forge")
    )
    def test_prereleases_etc_align_with_their_wildcarded_versions(self, version):
        assert sb._matches_version("1.20*", version)


class TestShulkerInstanceMatching:
    @staticmethod
    def matchall(shulker: ShulkerBox) -> list[str]:
        """Return the names of the testing instances that match the
        provided shulker

        Parameters
        ----------
        shulker : ShulkerBox
            The shulker to test

        Returns
        -------
        list of str
            The names of any testing instances that match the provided shulker
            (ordered in the order they're provided in utils.TESTING_INSTANCES)
        """
        return [
            instance.name
            for instance in utils.TESTING_INSTANCES
            if shulker.matches(instance)
        ]

    @pytest.mark.parametrize(
        "star_instances, star_tags",
        itertools.product(*((False, True),) * 2),
        ids=("", "tags=*", "instances=*", "instances=*,tags=*"),
    )
    def test_shulker_box_with_no_match_conditions_matches_everything(
        self, star_instances, star_tags
    ):
        match_criteria = []
        if star_instances:
            match_criteria.append(("instances", ("*",)))
        if star_tags:
            match_criteria.append(("tags", ("*",)))

        global_shulker = ShulkerBox(
            0, "global", Path("ignoreme"), tuple(match_criteria), ()
        )

        assert self.matchall(global_shulker) == [
            instance.name for instance in utils.TESTING_INSTANCES
        ]

    def test_matching_shulkers_by_instance_name(self):
        name_matching_shulker = ShulkerBox(
            0,
            "name_matching",
            Path("ignoreme"),
            (("instances", ("axolotl", "Chest Boat")),),
            (),
        )

        assert self.matchall(name_matching_shulker) == ["axolotl", "Chest Boat"]

    def test_instance_name_matching_is_exact(self):
        name_matching_shulker = ShulkerBox(
            0,
            "name_matching",
            Path("ignoreme"),
            (("instances", ("Bee", "Chest_Boat")),),
            (),
        )

        assert self.matchall(name_matching_shulker) == []

    def test_instance_name_matching_supports_wildcards(self):
        name_matching_shulker = ShulkerBox(
            0,
            "name_matching",
            Path("ignoreme"),
            (("instances", ("*est*",)),),
            (),
        )

        assert self.matchall(name_matching_shulker) == ["Chest Boat"]

    @pytest.mark.parametrize("quote_char", ('"', '"'), ids=("single", "double"))
    def test_instance_name_can_be_quoted(self, quote_char):
        name_matching_shulker = ShulkerBox(
            0,
            "name_matching",
            Path("ignoreme"),
            (("instances", (f"{quote_char}Chest Boat{quote_char}",)),),
            (),
        )

        assert self.matchall(name_matching_shulker) == ["Chest Boat"]

    @pytest.mark.parametrize(
        "quote_char", ("", '"', '"'), ids=("no-quote", "single-quote", "double-quote")
    )
    def test_instance_name_can_be_negated(self, quote_char):
        almost_global_shulker = ShulkerBox(
            0,
            "almost_global",
            Path("ignoreme"),
            (("instances", (f"!{quote_char}axolotl{quote_char}",)),),
            (),
        )

        assert self.matchall(almost_global_shulker) == [
            instance.name
            for instance in utils.TESTING_INSTANCES
            if instance.name != "axolotl"
        ]

    @pytest.mark.parametrize(
        "quote_char", ("", '"', '"'), ids=("no-quote", "single-quote", "double-quote")
    )
    def test_instance_name_negations_are_anded(self, quote_char):
        almost_global_shulker = ShulkerBox(
            0,
            "name_matching",
            Path("ignoreme"),
            (
                (
                    "instances",
                    (
                        f"!{quote_char}axolotl{quote_char}",
                        f"!{quote_char}Chest Boat{quote_char}",
                    ),
                ),
            ),
            (),
        )

        assert self.matchall(almost_global_shulker) == [
            instance.name
            for instance in utils.TESTING_INSTANCES
            if instance.name not in ("axolotl", "Chest Boat")
        ]

    @pytest.mark.parametrize(
        "quote_char", ('"', '"'), ids=("single-quote", "double-quote")
    )
    def test_instance_name_supports_regex(self, quote_char):
        pattern_matching_shulker = ShulkerBox(
            0,
            "pattern-matching",
            Path("ignoreme"),
            (
                (
                    "instances",
                    (f"r{quote_char}.+e.*{quote_char}",),
                ),
            ),
            (),
        )

        assert self.matchall(pattern_matching_shulker) == [
            "bee",
            "Chest Boat",
            "Drowned",
        ]

    def test_instance_name_matching_can_combine_patterns_and_negations(self):
        name_matching_shulker = ShulkerBox(
            0,
            "name_matching",
            Path("ignoreme"),
            (("instances", ("*e*", "!*oat")),),
            (),
        )

        assert self.matchall(name_matching_shulker) == ["bee", "Drowned"]

    def test_matching_shulkers_by_tag(self):
        tag_matching_shulker = ShulkerBox(
            0,
            "modz",
            Path("ignoreme"),
            (("tags", ("modded",)),),
            (),
        )

        assert self.matchall(tag_matching_shulker) == ["bee"]

    def test_tag_matching_ignores_case(self):
        tag_matching_shulker = ShulkerBox(
            0,
            "vanilla",
            Path("ignoreme"),
            (("tags", ("vanilla", "Vanilla Plus")),),
            (),
        )

        assert self.matchall(tag_matching_shulker) == [
            "official",
            "axolotl",
            "Chest Boat",
        ]

    def test_tag_matching_supports_wildcards(self):
        tag_matching_shulker = ShulkerBox(
            0,
            "vanilla-ish",
            Path("ignoreme"),
            (("tags", ("vanilla*",)),),
            (),
        )

        assert self.matchall(tag_matching_shulker) == [
            "official",
            "axolotl",
            "Chest Boat",
        ]

    def test_tag_matching_supports_negation(self):
        tag_matching_shulker = ShulkerBox(
            0,
            "vanilla-with-more-steps",
            Path("ignoreme"),
            (("tags", ("vanilla*", "!Vanilla Plus")),),
            (),
        )

        assert self.matchall(tag_matching_shulker) == [
            "official",
            "axolotl",
        ]

    def test_combining_instance_and_tag_negation(self):
        double_negative_shulker = ShulkerBox(
            0,
            "dont-wants",
            Path("ignoreme"),
            (
                ("instances", ("!Drowned",)),
                ("tags", ("!vanilla*",)),
            ),
            (),
        )
        assert self.matchall(double_negative_shulker) == ["bee"]

    def test_loader_matching_is_case_insensitive(self):
        # though in this case the values would have been normalized
        loader_matching_shulker = ShulkerBox(
            0,
            "forge instances",
            Path("ignoreme"),
            (("modloader", ("forge",)),),
            (),
        )

        assert self.matchall(loader_matching_shulker) == ["bee"]

    def test_loader_matching_treats_empty_string_as_vanilla(self):
        loader_matching_shulker = ShulkerBox(
            0,
            "Vanilla Only",
            Path("ignoreme"),
            (("modloader", ("",)),),
            (),
        )

        assert self.matchall(loader_matching_shulker) == ["official", "axolotl"]

    def test_loader_matching_accepts_wildcards(self):
        loader_matching_shulker = ShulkerBox(
            0,
            "Everybody",
            Path("ignoreme"),
            (("modloader", ("*",)),),
            (),
        )

        assert self.matchall(loader_matching_shulker) == [
            "official",
            "axolotl",
            "bee",
            "Chest Boat",
            "Drowned",
        ]

    def test_loader_matching_knows_how_to_interpret_fabric_like(self, tmp_path):
        (tmp_path / "shulker.cfg").write_text("[modloader]\nFabric-Like\n")

        loader_matching_shulker = ShulkerBox.from_cfg(tmp_path / "shulker.cfg")

        instances = [
            utils.instance("match_me", Path("foo"), modloader="Fabric Loader"),
            utils.instance("dont-match-me", Path("blah"), modloader="Fabulous Loader"),
            utils.instance("match me too", Path("bruh"), modloader="Quilt Loader"),
        ]

        assert [
            instance.name
            for instance in instances
            if loader_matching_shulker.matches(instance)
        ] == ["match_me", "match me too"]

    def test_loader_checks_multi_argument(self):
        loader_matching_shulker = ShulkerBox(
            0,
            "Fabric Instances",
            Path("ignoreme"),
            (
                (
                    "modloader",
                    (
                        "",
                        "*bric*",
                    ),
                ),
            ),
            (),
        )

        assert self.matchall(loader_matching_shulker) == [
            "official",
            "axolotl",
            "Chest Boat",
        ]

    def test_explicit_version_matching(self):
        minecraft_specific_shulker = ShulkerBox(
            0,
            "votey",
            Path("ignoreme"),
            (("minecraft", ("23w13a_or_b",)),),
            (),
        )

        assert self.matchall(minecraft_specific_shulker) == ["official"]

    @pytest.mark.parametrize(
        "version_spec",
        (
            ("1.19.0", "1.19.1", "1.19.2", "1.19.3", "1.19.4"),
            ("1.19",),
            ("1.19.*",),
            (">=1.19.0,<1.20",),
        ),
        ids=("explicit", "minor-version", "wildcard", "bounding"),
    )
    def test_version_bounding(self, version_spec):
        minecraft_specific_shulker = ShulkerBox(
            0,
            "votey",
            Path("ignoreme"),
            (("minecraft", version_spec),),
            (),
        )

        assert self.matchall(minecraft_specific_shulker) == ["official", "Chest Boat"]

    def test_instance_must_match_all_conditions(self):
        multi_condition_shulker = ShulkerBox(
            0,
            "finnicky",
            Path("ignoreme"),
            (
                ("modloader", ("Forge", "Fabric Loader")),
                ("tags", ("vanilla*",)),
                ("instances", ("official", "axolotl", "bee", "Chest Boat")),
            ),
            (),
        )
        assert self.matchall(multi_condition_shulker) == ["Chest Boat"]

    def test_unknown_condition_raises_error(self):
        whats_this_shulker = ShulkerBox(
            0,
            "out-of-left-field",
            Path("ignoreme"),
            (("quizzibuck", ("duketastrophe",)),),
            (),
        )
        with pytest.raises(
            NotImplementedError,
            match="Don't know how to apply match condition quizzibuck",
        ):
            self.matchall(whats_this_shulker)


@pytest.mark.usefixtures("multi_box_setup_teardown")
class TestMultiShulkerPlacing:
    @pytest.mark.parametrize("error_handling", ("ignore", "skip"))
    @pytest.mark.parametrize(
        "shulker_box, instance, should_match", utils.TESTING_SHULKER_INSTANCE_MATCHES
    )
    def test_multi_shulker_place_correctly_identifies_matches(
        self,
        minecraft_root,
        caplog,
        shulker_box,
        instance,
        should_match,
        error_handling,
    ):
        _ = place.place_ender_chest(minecraft_root, error_handling=error_handling)
        # TODO: test placements dict

        link_log = "\n".join(
            record.getMessage()
            for record in caplog.records
            if record.levelname == "INFO"
        )

        assert (
            f'{os.path.join(instance, ".minecraft")} to {shulker_box}' in link_log
        ) or (
            f'{os.path.join(instance, "minecraft")} to {shulker_box}' in link_log
        ) is should_match

    @pytest.mark.parametrize("error_handling", ("ignore", "skip"))
    def test_multi_shulker_place_overwrites_overlapping_symlinks(
        self, minecraft_root, error_handling
    ):
        resource_path = Path("resourcepacks") / "stuff.zip"
        placements = place.place_ender_chest(
            minecraft_root, error_handling=error_handling
        )

        assert placements["bee"][resource_path] == ["global", "optifine"]

        assert (
            minecraft_root / "instances" / "bee" / ".minecraft" / resource_path
        ).read_text() == "optifine-optimized!"

    def test_default_behavior_is_to_stop_at_first_error(self, minecraft_root, caplog):
        _ = place.place_ender_chest(minecraft_root)

        assert not (
            minecraft_root
            / "instances"
            / "chest_boat"
            / ".minecraft"
            / "mods"
            / "FoxNap.jar"
        ).exists()

        error_log = "\n".join(
            record.getMessage()
            for record in caplog.records
            if record.levelname == "ERROR"
        )

        assert error_log.endswith(
            f'{os.path.join("~", ".minecraft", "options.txt")}'  # fricking escape chars
            + " already exists\nAborting"
        )

    @pytest.mark.parametrize("prompt", (False, True), ids=["", "prompted"])
    def test_ignore_failures(
        self, home, minecraft_root, prompt, monkeypatch, caplog, capsys
    ):
        if prompt:
            monkeypatch.setattr("builtins.input", utils.scripted_prompt(("c",)))

        placements = place.place_ender_chest(
            minecraft_root,
            error_handling="prompt" if prompt else "ignore",
            relative=False,
        )
        _ = capsys.readouterr()  # suppress outputs

        assert placements["official"].get(Path("options.txt"), []) == []
        assert placements["official"][Path("mods") / "FoxNap.jar"] == ["1.19"]

        errors = [
            i for i, record in enumerate(caplog.records) if record.levelname == "ERROR"
        ]

        assert len(errors) == 1
        error_idx = errors[0]
        assert "options.txt already exists" in caplog.records[error_idx].getMessage()

        # note: there's actually no guarantee that this link didn't generate
        #       before the failure...
        assert (home / ".minecraft" / "mods" / "FoxNap.jar").is_symlink()  # and exists

    @pytest.mark.parametrize("prompt", (False, True), ids=["", "prompted"])
    def test_skip_match(
        self, home, minecraft_root, prompt, monkeypatch, caplog, capsys
    ):
        if prompt:
            monkeypatch.setattr("builtins.input", utils.scripted_prompt(("m",)))

        placements = place.place_ender_chest(
            minecraft_root, error_handling="prompt" if prompt else "skip"
        )
        _ = capsys.readouterr()  # suppress outputs

        assert placements["official"].get(Path("options.txt"), []) == []
        assert placements["official"][Path("data") / "achievements.txt"] == ["vanilla"]

        errors = [
            i for i, record in enumerate(caplog.records) if record.levelname == "ERROR"
        ]
        # meta-tests that I found the right line
        assert len(errors) == 1
        error_idx = errors[0]
        assert "options.txt already exists" in caplog.records[error_idx].getMessage()

        assert (
            caplog.records[error_idx + 1].levelname,
            caplog.records[error_idx + 1].msg,
        ) == ("WARNING", "Skipping the rest of this match")

        # and then make sure that it actually did move on

        assert caplog.records[error_idx + 2].levelname == "INFO"
        assert caplog.records[error_idx + 2].msg.startswith("Linking")

        # and now check for a subsequent place

        assert (
            home / ".minecraft" / "data" / "achievements.txt"
        ).read_text() == "Spelled acheivements correctly!"

    @pytest.mark.parametrize("prompt", (False, True), ids=["", "prompted"])
    def test_skip_instance(self, home, minecraft_root, prompt, monkeypatch, capsys):
        if prompt:
            monkeypatch.setattr("builtins.input", utils.scripted_prompt(("i",)))

        placements = place.place_ender_chest(
            minecraft_root, error_handling="prompt" if prompt else "skip-instance"
        )
        _ = capsys.readouterr()  # suppress outputs

        assert placements["official"].get(Path("options.txt"), []) == []
        assert placements["Chest Boat"][Path("options.txt")] == ["1.19"]

        assert (
            minecraft_root / "instances" / "chest-boat" / "minecraft" / "options.txt"
        ).read_text() == "autoJump:true"
        assert not (home / ".minecraft" / "data" / "achievements.txt").exists()

    @pytest.mark.parametrize("prompt", (False, True), ids=["", "prompted"])
    def test_skip_shulker_box(self, home, minecraft_root, prompt, monkeypatch, capsys):
        if prompt:
            monkeypatch.setattr("builtins.input", utils.scripted_prompt(("s",)))

        placements = place.place_ender_chest(
            minecraft_root, error_handling="prompt" if prompt else "skip-shulker"
        )
        _ = capsys.readouterr()  # suppress outputs

        assert placements["official"].get(Path("options.txt"), []) == []
        assert placements["Chest Boat"].get(Path("options.txt"), []) == []

        assert (
            home / ".minecraft" / "data" / "achievements.txt"
        ).read_text() == "Spelled acheivements correctly!"

        assert not (
            minecraft_root / "instances" / "chest-boat" / "minecraft" / "options.txt"
        ).exists()

    def test_raise_on_invalid_error_handling_arg(self, home, minecraft_root, caplog):
        with pytest.raises(ValueError, match="Unrecognized error-handling method"):
            _ = place.place_ender_chest(minecraft_root, error_handling="not a thing")

    def test_prompt_and_quit(self, home, minecraft_root, caplog, monkeypatch, capsys):
        monkeypatch.setattr("builtins.input", utils.scripted_prompt(("q",)))

        place.place_ender_chest(minecraft_root, error_handling="prompt")
        _ = capsys.readouterr()  # suppress outputs

        errors = [
            i for i, record in enumerate(caplog.records) if record.levelname == "ERROR"
        ]
        assert len(errors) == 2
        error_idx = errors[0]
        assert "options.txt already exists" in caplog.records[error_idx].getMessage()

        # and then make sure that it actually did abort

        assert not (home / ".minecraft" / "data" / "achievements.txt").exists()

    def test_prompt_and_retry(self, home, minecraft_root, caplog, monkeypatch) -> None:
        conflict_file_path = home / ".minecraft" / "options.txt"
        safe_keeping = minecraft_root / "options.txt.bkp"

        calls: list[str | None] = []

        def ope_lemme_delete_that(prompt: str | None = None) -> str:
            if calls:
                raise AssertionError("Should have only been called once")
            calls.append(prompt)
            conflict_file_path.rename(safe_keeping)
            return ""

        monkeypatch.setattr("builtins.input", ope_lemme_delete_that)

        try:
            placements = place.place_ender_chest(
                minecraft_root, error_handling="prompt", relative=False
            )
            assert placements["official"].get(Path("options.txt"), []) == ["1.19"]

            errors = [
                i
                for i, record in enumerate(caplog.records)
                if record.levelname == "ERROR"
            ]
            # meta-tests that I found the right line
            assert len(errors) == 1
            error_idx = errors[0]
            assert (
                "options.txt already exists" in caplog.records[error_idx].getMessage()
            )

            assert conflict_file_path.is_symlink()
        finally:
            if safe_keeping.exists():
                conflict_file_path.unlink(missing_ok=True)
                safe_keeping.rename(conflict_file_path)

    def test_invalid_prompt_response_will_just_keep_asking(
        self, home, minecraft_root, caplog, monkeypatch, capsys
    ):
        monkeypatch.setattr(
            "builtins.input",
            utils.scripted_prompt(("blargh", "grahr", "rrgh", "i dunno man", "q")),
        )

        _ = place.place_ender_chest(minecraft_root, error_handling="prompt")
        _ = capsys.readouterr()  # suppress outputs

        errors = [record for record in caplog.records if record.levelname == "ERROR"]
        assert len(errors) == 6

    def test_skip_shulker_box_that_doesnt_match_host(self, home, minecraft_root):
        with fs.shulker_box_config(minecraft_root, "1.19").open("a") as config_file:
            config_file.write(
                """
[hosts]
not-this-chest
"""
            )

        placements = place.place_ender_chest(minecraft_root)
        assert "1.19" not in set(sum(placements["Chest Boat"].values(), []))

        assert not (
            minecraft_root / "instances" / "chest-boat" / "minecraft" / "options.txt"
        ).exists()

        assert (
            home / ".minecraft" / "data" / "achievements.txt"
        ).read_text() == "Spelled acheivements correctly!"


@pytest.mark.usefixtures("multi_box_setup_teardown")
class TestPlacementCaching:
    def test_place_cache_roundtrip(self, home, minecraft_root, caplog):
        placements = place.place_ender_chest(minecraft_root, error_handling="ignore")
        place.cache_placements(minecraft_root, placements)
        assert placements == place.load_placement_cache(minecraft_root)


@pytest.fixture
def placement_cache(home, minecraft_root, caplog, multi_box_setup_teardown):
    placements = place.place_ender_chest(
        minecraft_root, error_handling="ignore", relative=False
    )
    place.cache_placements(minecraft_root, placements)
    yield place.load_placement_cache(minecraft_root)
    fs.place_cache(minecraft_root).unlink()


class TestResourceTracing:
    @pytest.mark.parametrize(
        "pattern_type",
        ("filename", "resource_path", "glob", "relative_path", "absolute_path"),
    )
    def test_track_down_a_simple_single_file(
        self, placement_cache, minecraft_root, pattern_type
    ):
        instance_folder = Path("instances") / "bee" / ".minecraft"
        resource_path = Path("mods") / "optifine.jar"

        pattern = str(
            {
                "filename": resource_path.name,
                "resource_path": resource_path,
                "glob": "opti*.jar",
                "relative_path": instance_folder / resource_path,
                "absolute_path": (
                    minecraft_root / instance_folder / resource_path
                ).absolute(),
            }[pattern_type]
        )
        assert place.trace_resource(
            minecraft_root,
            pattern,
            instance_name="bee",
            placements=placement_cache,
        ) == [(instance_folder, resource_path, ["optifine"])]

    @pytest.mark.parametrize(
        "pattern_type",
        ("filename", "resource_path", "glob", "relative_path", "absolute_path"),
    )
    def test_track_down_a_single_multi_linked_file(
        self, placement_cache, minecraft_root, pattern_type
    ):
        instance_folder = Path("instances") / "bee" / ".minecraft"
        resource_path = Path("resourcepacks") / "stuff.zip"

        pattern = str(
            {
                "filename": resource_path.name,
                "resource_path": resource_path,
                "glob": "stuff*",
                "relative_path": instance_folder / resource_path,
                "absolute_path": (
                    minecraft_root / instance_folder / resource_path
                ).absolute(),
            }[pattern_type]
        )
        assert place.trace_resource(
            minecraft_root,
            pattern,
            instance_name="bee",
            placements=placement_cache,
        ) == [(instance_folder, resource_path, ["global", "optifine"])]

    def test_track_down_multiple_files(self, placement_cache, minecraft_root):
        instance_folder = Path("instances") / "bee" / ".minecraft"

        assert sorted(
            place.trace_resource(
                minecraft_root,
                "*.zip",
                instance_name="bee",
                placements=placement_cache,
            ),
            key=lambda x: str(x[0] / x[1]).lower(),
        ) == [
            (
                instance_folder,
                Path("resourcepacks") / "stuff.zip",
                ["global", "optifine"],
            ),
            (instance_folder, Path("resourcepacks") / "TEAVSRP.zip", ["global"]),
            (instance_folder, Path("shaderpacks") / "Seuss CitH.zip", ["optifine"]),
        ]

    def test_search_all_instances(self, placement_cache, minecraft_root):
        instance_folders = {
            instance.name: instance.root
            for instance in inventory.load_ender_chest_instances(
                minecraft_root, log_level=logging.DEBUG
            )
        }

        mods_folder = Path("mods")

        assert sorted(
            place.trace_resource(
                minecraft_root,
                "*.jar",
                placements=placement_cache,
            ),
            key=lambda x: str(x[0] / x[1]).lower(),
        ) == [
            (instance_folders["bee"], mods_folder / "BME.jar", ["optifine"]),
            (instance_folders["bee"], mods_folder / "optifine.jar", ["optifine"]),
            (instance_folders["Chest Boat"], mods_folder / "FoxNap.jar", ["1.19"]),
            (instance_folders["official"], mods_folder / "FoxNap.jar", ["1.19"]),
        ]


@pytest.mark.usefixtures("multi_box_setup_teardown")
class TestListPlacements:
    @pytest.mark.parametrize("fail_type", ("no_cache", "bad_cache"))
    def test_if_no_cache_then_suggest_running_place(
        self, minecraft_root, fail_type, caplog
    ):
        if fail_type == "bad_cache":
            fs.place_cache(minecraft_root).write_text("This is not a json")

        place.list_placements(minecraft_root, pattern="*")

        errors = [record for record in caplog.records if record.levelname == "ERROR"]

        assert len(errors) == 1
        assert "enderchest place" in errors[0].message

    @pytest.mark.usefixtures("placement_cache")
    def test_errors_out_if_no_such_instance(self, minecraft_root, caplog):
        place.list_placements(minecraft_root, pattern="*", instance_name="utopia")

        errors = [record for record in caplog.records if record.levelname == "ERROR"]

        assert len(errors) == 1
        assert "No instance named" in errors[0].message

    @pytest.mark.usefixtures("placement_cache")
    def test_warns_if_no_matches(self, minecraft_root, caplog):
        place.list_placements(minecraft_root, pattern="nemo")

        problems = [
            record
            for record in caplog.records
            if record.levelname in ("WARNING", "ERROR")
        ]

        assert len(problems) == 1
        assert problems[0].levelname == "WARNING"
        assert (
            "Note: this command does not check inside linked folders."
            in problems[0].message
        )

    @pytest.mark.usefixtures("placement_cache")
    def test_only_top_priority_box_is_important(self, minecraft_root, caplog):
        place.list_placements(minecraft_root, pattern="stuff.zip", instance_name="bee")

        logs = defaultdict(list)
        for record in caplog.records:
            logs[record.levelno].append(record)

        assert len(logs[logging.ERROR]) == len(logs[logging.WARNING]) == 0

        assert len(logs[IMPORTANT]) == 2 and len(logs[logging.INFO]) == 1
        assert all(
            (
                "resolves to" in logs[IMPORTANT][0].message,
                "optifine" in logs[IMPORTANT][1].args,
                "global" in logs[logging.INFO][0].args,
            )
        )

    def test_warns_if_a_symlink_no_longer_exists(
        self, minecraft_root, caplog, placement_cache
    ):
        placement_cache["Chest Boat"][Path("config") / "BME" / "options.txt"] = [
            "optifine"
        ]
        place.cache_placements(minecraft_root, placement_cache)

        place.list_placements(
            minecraft_root, pattern="options.txt", instance_name="Chest Boat"
        )

        logs = defaultdict(list)
        for record in caplog.records:
            logs[record.levelno].append(record)

        assert len(logs[logging.ERROR]) == 0 and len(logs[logging.INFO]) == 0

        assert len(logs[logging.WARNING]) == 1 and len(logs[IMPORTANT]) == 3

        assert all(
            (
                "no longer exists" in logs[logging.WARNING][0].message,
                "optifine" in logs[IMPORTANT][0].args,
                "resolves to" in logs[IMPORTANT][1].message,
                "1.19" in logs[IMPORTANT][2].args,
            )
        ) or all(  # doesn't matter which comes first
            (
                "resolves to" in logs[IMPORTANT][0].message,
                "1.19" in logs[IMPORTANT][1].args,
                "no longer exists" in logs[logging.WARNING][0].message,
                "optifine" in logs[IMPORTANT][2].args,
            )
        )



"""Specification and configuration of an EnderChest"""

from collections.abc import Iterable
from dataclasses import dataclass
from pathlib import Path
from socket import gethostname
from typing import Any
from urllib.parse import ParseResult, urlparse

from . import config as cfg
from . import filesystem as fs
from . import instance as i
from . import sync
from .loggers import CRAFT_LOGGER, GATHER_LOGGER, INVENTORY_LOGGER
from .sync import abspath_from_uri

_DEFAULTS = (
    ("offer_to_update_symlink_allowlist", True),
    ("sync_confirm_wait", 5),
    ("place_after_open", True),
    ("do_not_sync", ("EnderChest/enderchest.cfg", "EnderChest/.*", ".DS_Store")),
    (
        "shulker_box_folders",
        (
            "config",
            "mods",
            "resourcepacks",
            "saves",
            "shaderpacks",
        ),
    ),
    ("standard_link_folders", ()),
    (
        "global_link_folders",
        (
            "backups",
            "cachedImages",
            "crash-reports",
            "logs",
            "replay_recordings",
            "screenshots",
            "schematics",
            "config/litematica",  # still worth having in case max_depth>2
            ".bobby",
        ),
    ),
)


@dataclass(init=False, repr=False, eq=False)
class EnderChest:
    """Configuration of an EnderChest

    Parameters
    ----------
    uri : URI or Path
        The "address" of this EnderChest, ideally as it can be accessed from other
        EnderChest installations, including both the path to where
        the EnderChest folder can be found (that is, the parent of the
        EnderChest folder itself, aka the "minecraft_root"), its net location
        including credentials, and the protocol that should be used to perform
        the syncing. All that being said, if just a path is provided, the
        constructor will try to figure out the rest.
    name : str, optional
        A unique name to give to this EnderChest installation. If None is
        provided, this will be taken from the hostname of the supplied URI.
    instances : list-like of InstanceSpec, optional
        The list of instances to register with this EnderChest installation
    remotes : list-like of URI, or (URI, str) tuples
        A list of other installations that this EnderChest should be aware of
        (for syncing purposes). When a (URI, str) tuple is provided, the
        second value will be used as the name/alias of the remote.

    Attributes
    ----------
    name : str
        The unique name of this EnderChest installation. This is most commonly
        the computer's hostname, but one can configure multiple EnderChests
        to coexist on the same system (either for the sake of having a "cold"
        backup or for multi-user systems).
    uri : str
        The complete URI of this instance
    root : Path
        The path to this EnderChest folder
    instances : list-like of InstanceSpec
        The instances registered with this EnderChest
    remotes : list-like of (ParseResult, str) pairs
        The other EnderChest installations this EnderChest is aware of, paired
        with their aliases
    offer_to_update_symlink_allowlist : bool
        By default, EnderChest will offer to create or update `allowed_symlinks.txt`
        on any 1.20+ instances that do not already blanket allow links into
        EnderChest. **EnderChest will never modify that or any other Minecraft
        file without your express consent.** If you would prefer to edit these
        files yourself (or simply not symlink your world saves), change this
        parameter to False.
    sync_confirm_wait : bool or int
        The default behavior when syncing EnderChests is to first perform a dry
        run of every sync operation and then wait 5 seconds before proceeding with the
        real sync. The idea is to give the user time to interrupt the sync if
        the dry run looks wrong. This can be changed by either raising or lowering
        the value of confirm, by disabling the dry-run-first behavior entirely
        (`confirm=False`) or by requiring that the user explicitly confirms
        the sync (`confirm=True`). This default behavior can also be overridden
        when actually calling the sync commands.
    place_after_open: bool
        By default, EnderChest will follow up any `enderchest open` operation
        with an `enderchest place` to refresh any changed symlinks. This
        functionality can be disabled by setting this parameter to False.
    do_not_sync : list of str
        Glob patterns of files that should not be synced between EnderChest
        installations. By default, this list comprises `EnderChest/enderchest.cfg`,
        any top-level folders starting with a "." (like .git) and
        `.DS_Store` (for all you mac gamers).
    shulker_box_folders : list of str
        The folders that will be created inside each new shulker box
    standard_link_folders : list of str
        The default set of "link folders" when crafting a new shulker box
    global_link_folders : list of str
        The "global" set of "link folders," offered as a suggestion when
        crafting a new shulker box
    """

    name: str
    _uri: ParseResult
    _instances: list[i.InstanceSpec]
    _remotes: dict[str, ParseResult]
    offer_to_update_symlink_allowlist: bool
    sync_confirm_wait: bool | int
    place_after_open: bool
    do_not_sync: list[str]
    shulker_box_folders: list[str]
    standard_link_folders: list[str]
    global_link_folders: list[str]

    def __init__(
        self,
        uri: str | ParseResult | Path,
        name: str | None = None,
        remotes: (
            Iterable[str | ParseResult | tuple[str, str] | tuple[ParseResult, str]]
            | None
        ) = None,
        instances: Iterable[i.InstanceSpec] | None = None,
    ):
        for setting, value in _DEFAULTS:
            setattr(self, setting, list(value) if isinstance(value, tuple) else value)

        try:
            if isinstance(uri, ParseResult):
                self._uri = uri
            elif isinstance(uri, Path):
                self._uri = urlparse(uri.absolute().as_uri())
            else:
                self._uri = urlparse(uri)
        except AttributeError as parse_problem:  # pragma: no cover
            raise ValueError(f"{uri} is not a valid URI") from parse_problem

        if not self._uri.netloc:
            self._uri = self._uri._replace(netloc=sync.get_default_netloc())
        if not self._uri.scheme:
            self._uri = self._uri._replace(scheme=sync.DEFAULT_PROTOCOL)

        self.name = name or self._uri.hostname or gethostname()

        self._instances = []
        self._remotes = {}

        for instance in instances or ():
            self.register_instance(instance)

        for remote in remotes or ():
            if isinstance(remote, (str, ParseResult)):
                self.register_remote(remote)
            else:
                self.register_remote(*remote)

    @property
    def uri(self) -> str:
        return self._uri.geturl()

    def __repr__(self) -> str:
        return f"EnderChest({self.uri, self.name})"

    @property
    def root(self) -> Path:
        return fs.ender_chest_folder(abspath_from_uri(self._uri), check_exists=False)

    @property
    def instances(self) -> tuple[i.InstanceSpec, ...]:
        return tuple(self._instances)

    def register_instance(self, instance: i.InstanceSpec) -> i.InstanceSpec:
        """Register a new Minecraft installation

        Parameters
        ----------
        instance : InstanceSpec
            The instance to register

        Returns
        -------
        InstanceSpec
            The spec of the instance as it was actually registered (in case the
            name changed or somesuch)

        Notes
        -----
        - If the instance's name is already assigned to a registered instance,
          this method will choose a new one
        - If this instance shares a path with an existing instance, it will
          replace that instance
        """
        matching_instances: list[i.InstanceSpec] = []
        for old_instance in self._instances:
            if i.equals(abspath_from_uri(self._uri), instance, old_instance):
                matching_instances.append(old_instance)
                self._instances.remove(old_instance)

        instance = i.merge(*matching_instances, instance)

        name = instance.name
        counter = 0
        taken_names = {old_instance.name for old_instance in self._instances}
        while True:
            if name not in taken_names:
                break
            counter += 1
            name = f"{instance.name}.{counter}"

        GATHER_LOGGER.debug("Registering instance %s at %s", name, instance.root)
        self._instances.append(instance._replace(name=name))
        return self._instances[-1]

    @property
    def remotes(self) -> tuple[tuple[ParseResult, str], ...]:
        return tuple((remote, alias) for alias, remote in self._remotes.items())

    def register_remote(
        self, remote: str | ParseResult, alias: str | None = None
    ) -> None:
        """Register a new remote EnderChest installation (or update an existing
        registry)

        Parameters
        ----------
        remote : URI
            The URI of the remote
        alias : str, optional
            an alias to give to this remote. If None is provided, the URI's hostname
            will be used.

        Raises
        ------
        ValueError
            If the provided remote is invalid
        """
        try:
            remote = remote if isinstance(remote, ParseResult) else urlparse(remote)
            alias = alias or remote.hostname
            if not alias:  # pragma: no cover
                raise AttributeError(f"{remote.geturl()} has no hostname")
            GATHER_LOGGER.debug("Registering remote %s (%s)", remote.geturl(), alias)
            self._remotes[alias] = remote
        except AttributeError as parse_problem:  # pragma: no cover
            raise ValueError(f"{remote} is not a valid URI") from parse_problem

    @classmethod
    def from_cfg(cls, config_file: Path) -> "EnderChest":
        """Parse an EnderChest from its config file

        Parameters
        ----------
        config_file : Path
            The path to the config file

        Returns
        -------
        EnderChest
            The resulting EnderChest

        Raises
        ------
        ValueError
            If the config file at that location cannot be parsed
        FileNotFoundError
            If there is no config file at the specified location
        """
        INVENTORY_LOGGER.debug("Reading config file from %s", config_file)
        config = cfg.read_cfg(config_file)

        # All I'm gonna say is that Windows pathing is the worst
        path = urlparse(config_file.absolute().parent.parent.as_uri()).path

        instances: list[i.InstanceSpec] = []
        remotes: list[str | tuple[str, str]] = []

        requires_rewrite = False

        scheme: str | None = None
        netloc: str | None = None
        name: str | None = None
        sync_confirm_wait: str | None = None
        place_after_open: bool | None = None
        offer_to_update_symlink_allowlist: bool = True
        do_not_sync: list[str] | None = None
        folder_defaults: dict[str, list[str] | None] = {
            "shulker_box_folders": None,
            "standard_link_folders": None,
            "global_link_folders": None,
        }

        for section in config.sections():
            if section == "properties":
                scheme = config[section].get("sync-protocol")
                netloc = config[section].get("address")
                name = config[section].get("name")
                sync_confirm_wait = config[section].get("sync-confirmation-time")
                place_after_open = config[section].getboolean("place-after-open")
                offer_to_update_symlink_allowlist = config[section].getboolean(
                    "offer-to-update-symlink-allowlist", True
                )
                if "do-not-sync" in config[section]:
                    do_not_sync = cfg.parse_ini_list(
                        config[section]["do-not-sync"] or ""
                    )
                for setting in folder_defaults:
                    setting_key = setting.replace("_", "-")
                    if setting_key in config[section]:
                        folder_defaults[setting] = cfg.parse_ini_list(
                            config[section][setting_key] or ""
                        )
            elif section == "remotes":
                for remote in config[section].items():
                    if remote[1] is None:
                        raise ValueError("All remotes must have an alias specified")
                    remotes.append((remote[1], remote[0]))
            else:
                # TODO: flag requires_rewrite if instance was normalized
                instances.append(i.InstanceSpec.from_cfg(config[section]))

        scheme = scheme or sync.DEFAULT_PROTOCOL
        netloc = netloc or sync.get_default_netloc()
        uri = ParseResult(
            scheme=scheme, netloc=netloc, path=path, params="", query="", fragment=""
        )

        ender_chest = EnderChest(uri, name, remotes, instances)
        if sync_confirm_wait is not None:
            match sync_confirm_wait.lower():
                case "true" | "prompt" | "yes" | "confirm":
                    ender_chest.sync_confirm_wait = True
                case "false" | "no" | "skip":
                    ender_chest.sync_confirm_wait = False
                case _:
                    try:
                        ender_chest.sync_confirm_wait = int(sync_confirm_wait)
                    except ValueError as bad_input:
                        raise ValueError(
                            "Invalid value for sync-confirmation-time:"
                            f" {sync_confirm_wait}"
                        ) from bad_input
        if place_after_open is None:
            INVENTORY_LOGGER.warning(
                "This EnderChest does not have a value set for place-after-open."
                "\nIt is being set to False for now. To enable this functionality,"
                "\nedit the value in %s",
                config_file,
            )
            ender_chest.place_after_open = False
            requires_rewrite = True
        else:
            ender_chest.place_after_open = place_after_open

        ender_chest.offer_to_update_symlink_allowlist = (
            offer_to_update_symlink_allowlist
        )

        if do_not_sync is not None:
            ender_chest.do_not_sync = do_not_sync
            chest_cfg_exclusion = "/".join(
                (fs.ENDER_CHEST_FOLDER_NAME, fs.ENDER_CHEST_CONFIG_NAME)
            )
            if chest_cfg_exclusion not in do_not_sync:
                INVENTORY_LOGGER.warning(
                    "This EnderChest was not configured to exclude the EnderChest"
                    " config file from sync operations."
                    "\nThat is being fixed now."
                )
                ender_chest.do_not_sync.insert(0, chest_cfg_exclusion)
                requires_rewrite = True
        for setting in folder_defaults:
            if folder_defaults[setting] is None:
                folder_defaults[setting] = dict(_DEFAULTS)[setting]  # type: ignore
                # requires_rewrite = True  # though I'm considering it
            setattr(ender_chest, setting, folder_defaults[setting])

        if requires_rewrite:
            ender_chest.write_to_cfg(config_file)
            return cls.from_cfg(config_file)
        return ender_chest

    def write_to_cfg(self, config_file: Path | None = None) -> str:
        """Write this EnderChest's configuration to INI

        Parameters
        ----------
        config_file : Path, optional
            The path to the config file, assuming you'd like to write the
            contents to file

        Returns
        -------
        str
            An INI-syntax rendering of this EnderChest's config

        Notes
        -----
        The "root" attribute is ignored for this method
        """
        properties: dict[str, Any] = {
            "name": self.name,
            "address": self._uri.netloc,
            "sync-protocol": self._uri.scheme,
        }
        if self.sync_confirm_wait is True:
            properties["sync-confirmation-time"] = "prompt"
        else:
            properties["sync-confirmation-time"] = self.sync_confirm_wait

        for setting, _ in _DEFAULTS:
            if setting == "sync_confirm_wait":
                continue  # already did this one
            setting_key = setting.replace("_", "-")
            properties[setting_key] = getattr(self, setting)

        remotes: dict[str, str] = {name: uri.geturl() for uri, name in self.remotes}

        instances: dict[str, dict[str, Any]] = {}

        for instance in self.instances:
            instances[instance.name] = {
                "root": instance.root,
                "minecraft-version": instance.minecraft_versions,
                "modloader": instance.modloader,
                "groups": instance.groups_,
                "tags": instance.tags_,
            }

        config = cfg.dumps(
            fs.ENDER_CHEST_CONFIG_NAME, properties, remotes=remotes, **instances
        )

        if config_file:
            CRAFT_LOGGER.debug("Writing configuration file to %s", config_file)
            config_file.write_text(config)
        return config


def create_ender_chest(minecraft_root: Path, ender_chest: EnderChest) -> None:
    """Create an EnderChest based on the provided configuration

    Parameters
    ----------
    minecraft_root : Path
        The root directory that your minecraft stuff is in (or, at least, the
        one inside which you want to create your EnderChest)
    ender_chest : EnderChest
        The spec of the chest to create

    Notes
    -----
    - The "root" attribute of the EnderChest config will be ignored--instead
      the EnderChest will be created at <minecraft_root>/EnderChest
    - This method does not check to see if there is already an EnderChest set
      up at the specified location--if one exists, its config will
      be overwritten
    """
    root = fs.ender_chest_folder(minecraft_root, check_exists=False)
    root.mkdir(exist_ok=True)

    config_path = fs.ender_chest_config(minecraft_root, check_exists=False)
    ender_chest.write_to_cfg(config_path)
    CRAFT_LOGGER.info("EnderChest configuration written to %s", config_path)



"""Functionality for managing the EnderChest and shulker box config files and folders"""

import itertools
import os
from collections.abc import Iterable
from pathlib import Path

from .loggers import INVENTORY_LOGGER

ENDER_CHEST_FOLDER_NAME = "EnderChest"

ENDER_CHEST_CONFIG_NAME = "enderchest.cfg"

SHULKER_BOX_CONFIG_NAME = "shulkerbox.cfg"

PLACE_CACHE_NAME = ".place_cache.json"


def ender_chest_folder(minecraft_root: Path, check_exists: bool = True) -> Path:
    """Given a minecraft root directory, return the path to the EnderChest
    folder

    Parameters
    ----------
    minecraft_root : Path
        The root directory that your minecraft stuff (or, at least, the one
        that's the parent of your EnderChest folder)
    check_exists : bool, optional
        By default, this method will raise an error if no EnderChest exists
        at that location (meaning no folder or no enderchest config file in
        that folder). To disable that check, call this method with
        `check_exists=False`.

    Returns
    -------
    Path
        The path to the EnderChest folder

    Raises
    ------
    FileNotFoundError
        If no valid EnderChest installation exists within the given
        minecraft root (and checking hasn't been disabled)
    """
    return ender_chest_config(minecraft_root, check_exists=check_exists).parent


def ender_chest_config(minecraft_root, check_exists: bool = True) -> Path:
    """Given a minecraft root directory, return the path to the EnderChest
    config file

    Parameters
    ----------
    minecraft_root : Path
        The root directory that your minecraft stuff (or, at least, the one
        that's the parent of your EnderChest folder)
    check_exists : bool, optional
        By default, this method will raise an error if the enderchest config
        file does not already exist. To disable that check, call this method
        with `check_exists=False`.

    Returns
    -------
    Path
        The path to the EnderChest config file

    Raises
    ------
    FileNotFoundError
        If the EnderChest config file isn't where it's supposed to be (and
        checking hasn't been disabled)

    Notes
    -----
    This method does not check if the config file is valid
    """
    config_path = minecraft_root / ENDER_CHEST_FOLDER_NAME / ENDER_CHEST_CONFIG_NAME

    if check_exists and not config_path.exists():
        raise FileNotFoundError(
            f"No valid EnderChest installation exists within {minecraft_root}"
        )
    return config_path


def shulker_box_root(minecraft_root: Path, shulker_box_name: str) -> Path:
    """Generate the path to the root of a shulker box, given its name and the
    minecraft root directory

    Parameters
    ----------
    minecraft_root : Path
        The root directory that your minecraft stuff (or, at least, the one
        that's the parent of your EnderChest folder)
    shulker_box_name : str
        The name of the shulker box to resolve

    Returns
    -------
    Path
        The path to the shulker box folder

    Notes
    -----
    This method does not check if a shulker box exists at that location
    """
    return ender_chest_folder(minecraft_root) / shulker_box_name


def shulker_box_config(minecraft_root: Path, shulker_box_name: str) -> Path:
    """Generate the path to a shulker box config file, given its name and
    the minecraft root directory

    Parameters
    ----------
    minecraft_root : Path
        The root directory that your minecraft stuff (or, at least, the one
        that's the parent of your EnderChest folder)
    shulker_box_name : str
        The name of the shulker box to resolve

    Returns
    -------
    Path
        The path to the shulker box folder

    Notes
    -----
    This method does not check if a shulker box config exists at that location
    """
    return shulker_box_root(minecraft_root, shulker_box_name) / SHULKER_BOX_CONFIG_NAME


def place_cache(minecraft_root: Path) -> Path:
    """Given a minecraft root directory, return the path to the EnderChest
    place cache file

    Parameters
    ----------
    minecraft_root : Path
        The root directory that your minecraft stuff (or, at least, the one
        that's the parent of your EnderChest folder)

    Returns
    -------
    Path
        The path to the place cache file

    Notes
    -----
    This method does not check if the cache file is valid or if it even exists
    """
    return ender_chest_folder(minecraft_root) / PLACE_CACHE_NAME


def shulker_box_configs(minecraft_root: Path) -> Iterable[Path]:
    """Find all shulker box configs on the system

    Parameters
    ----------
    minecraft_root : Path
        The root directory that your minecraft stuff (or, at least, the one
        that's the parent of your EnderChest folder)

    Returns
    -------
    list-like of paths
        The paths to all the shulker box configs on the system

    Notes
    -----
    This method does not check to make sure those config files are valid,
    just that they exist
    """
    INVENTORY_LOGGER.debug("Searching for shulker configs within %s", minecraft_root)
    return ender_chest_folder(minecraft_root).glob(f"*/{SHULKER_BOX_CONFIG_NAME}")


def minecraft_folders(search_path: Path) -> Iterable[Path]:
    """Find all .minecraft folders within a given search path

    Parameters
    ----------
    search_path : Path
        The directory to search

    Returns
    -------
    list-like of paths
        The paths to all the .minecraft folders this method could find

    Notes
    -----
    This method does not check to make sure that those .minecraft folders
    contain valid minecraft instances, just that they exist
    """
    return itertools.chain(
        search_path.rglob(".minecraft"), search_path.rglob("minecraft")
    )


def links_into_enderchest(
    minecraft_root: Path, link: Path, check_exists: bool = True
) -> bool:
    """Determine whether a symlink's target is inside the EnderChest specified
    by the Minecraft root.

    Parameters
    ----------
    minecraft_root : Path
        The root directory that your minecraft stuff (or, at least, the one
        that's the parent of your EnderChest folder)
    link : Path
        The link to check
    check_exists : bool, optional
        By default, this method will raise an error if no EnderChest exists
        at that location (meaning no folder or no enderchest config file in
        that folder). To disable that check, call this method with
        `check_exists=False`.

    Returns
    -------
    bool
        True if the path is inside of the EnderChest folder. False otherwise.

    Notes
    -----
    This method only checks the *direct target* of the link as opposed to the
    fully resolved path.

    Raises
    ------
    FileNotFoundError
        If no valid EnderChest installation exists within the given
        minecraft root (and checking hasn't been disabled)
    OSError
        If the link provided isn't actually a symbolic link
    """
    chest_folder = os.path.normpath(
        ender_chest_folder(minecraft_root, check_exists=check_exists)
        .expanduser()
        .absolute()
    )

    target = os.readlink(link)
    if not os.path.isabs(target):
        target = os.path.normpath(link.parent / target)

    # Windows shenanigans: https://bugs.python.org/issue42957
    if target.startswith(("\\\\?\\", "\\??\\")):  # pragma: no cover
        try:
            os.stat(target[4:])
            target = target[4:]
        except (OSError, FileNotFoundError):
            # then maybe this is somehow legit
            pass

    # there's probably a better way to check if a file is inside a sub-path
    try:
        common_root = os.path.commonpath([target, chest_folder])
    except ValueError:  # if they have no common root
        common_root = ""
    return common_root == chest_folder
